# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import verifytweet.services.image as image_service
import verifytweet.services.text as text_service
import verifytweet.util.validator as validator

from verifytweet.util.logging import logger
from verifytweet.util.result import ResultStatus


def extract_and_parse(file_path: str):
    """Preprocess text from image

    Extracts text from image using image service,
    parses entities from text using text service.

    Args:
        file_path: represents path of the image file.

    Returns:
        entities: Entities parsed from text such as tweet, user_id and date.
        status: Enum ResultStatus representing result status

    """
    if not isinstance(file_path, str):
            raise TypeError('File path must be type string')
    if not file_path:
        raise ValueError('File path must be a valid path')
    try:
        text_extractor = image_service.Extractor()
        extracted_text, extractor_status = text_extractor.get_text(file_path)
    except Exception as e:
        logger.exception(e)
        return (None, ResultStatus.MODULE_FAILURE)
    if extractor_status != ResultStatus.ALL_OKAY:
        return (None, extractor_status)
    logger.debug('Processed text: ' + extracted_text)

    try:
        entity_parser = text_service.DataParser()
        entities, parser_status = entity_parser.get_entities(extracted_text)
    except Exception as e:
        logger.exception(e)
        return (None, ResultStatus.MODULE_FAILURE)
    if parser_status != ResultStatus.ALL_OKAY:
        return (None, parser_status)
    logger.debug('Entities: ' + str(entities))
    return (entities, parser_status)


def calculate_and_validate(entities: dict, tweet_text_list: list):
    """Calculates similarity matrix and validates tweet

    Calculates a similarity matrix from same day tweet
    corpus using text service and validates tweet
    using validator

    Args:
        entities: represents dictionary of entities extracted from text
        tweet_text_list: list of strings representing same day tweets

    Returns:
        valid_tweet: Validity status of tweet
        status: Enum ResultStatus representing result status

    """
    try:
        text_processor = text_service.TextProcessor()
        similarity_matrix, processor_status = text_processor.get_similarity(
            entities['tweet'], tweet_text_list)
    except Exception as e:
        logger.exception(e)
        return (None, None, ResultStatus.MODULE_FAILURE)
    if processor_status != ResultStatus.ALL_OKAY:
        return (None, None, processor_status)

    try:
        valid_tweet, match_index, validator_status = validator.verify_validity(
            similarity_matrix)
    except Exception as e:
        logger.exception(e)
        return (None, None, ResultStatus.MODULE_FAILURE)
    if validator_status == ResultStatus.MODULE_FAILURE:
        return (None, None, validator_status)
    logger.debug('Tweet Validity: ' + str(valid_tweet))
    if not valid_tweet:
        return (False, None, ResultStatus.NO_RESULT)
    return (valid_tweet, match_index-1, ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import multiprocessing
import tempfile
import logging


def no_of_workers():
    """Returns no.of workers to be spawned

    Retrieves cpu count using multiprocessing module.
    No.of workers suggested is one more than 
    twice the cpu count.

    """
    return multiprocessing.cpu_count() * 2 + 1


class Config(object):
    """Configuration for the app
    """

    IMAGEMAGICK_PATH = os.getenv('IMAGEMAGICK_PATH') if os.getenv(
        'IMAGEMAGICK_PATH') else "convert"
    FILE_DIRECTORY = tempfile.mkdtemp()
    TWEET_MAX_STORE = 150
    RUN_METHOD = "cli"
    LOG_LEVEL = logging.DEBUG if os.getenv('VERIFYTWEET_DEBUG') else logging.INFO
    ALLOWED_EXTENSIONS = set(["png", "jpg", "jpeg"])
    SIMILARITY_THRESHOLD = 0.6


class TwitterAPIConfig(Config):
    """Twitter API Configuration
    """

    TWITTER_ACCESSTOKEN = os.getenv('TWITTER_ACCESSTOKEN')
    TWITTER_HOSTNAME = "https://api.twitter.com"
    TWITTER_APIVER = "1.1"
    TWITTER_CONTEXT = "statuses/user_timeline.json"
    TWEET_DATE_KEY = "created_at"
    TWEET_USERNAME_KEY = "screen_name"
    TWEET_COUNT = 150
    TWEET_COUNT_KEY = "count"
    TWEET_MAX_OLD = 7
    TWEET_TEXT_KEY = "text"


class WebConfig(Config):
    """Configuration for running as web service
    """

    APP_HOST = os.getenv('APP_HOST') or '0.0.0.0'
    APP_PORT = os.getenv('APP_PORT') or '8000'
    TIMEOUT = 2000
    MAX_CONTENT_LENGTH_KEY = "MAX_CONTENT_LENGTH"
    MAX_CONTENT_LENGTH = 2097152
    WORKER_COUNT = no_of_workers()
    WORKER_CLASS = "eventlet"


run_method = "web" if "VERIFYTWEET_RUN_FOR_WEB" in os.environ else "cli"
Config.RUN_METHOD = run_method
configurations = dict({
    "cli": Config,
    "web": WebConfig,
})

app_config = configurations[run_method]



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy

from verifytweet import validator
from verifytweet import settings
from verifytweet import result

app_config = settings.app_config
app_config.SIMILARITY_THRESHOLD = 0.6

def test_validator_empty_input():
    """Test verify validity for empty input
    """
    with pytest.raises(TypeError):
        validator.verify_validity()


def test_validator_invalid_type_input():
    """Test verify validity for invalid type input
    """
    with pytest.raises(TypeError):
        validator.verify_validity(None)
        validator.verify_validity(list())
        validator.verify_validity([[]])


def test_validator_valid_similarity_matrix():
    """Test verfiy validity for valid similarity matrix
    """
    test_numpy_array = numpy.array([[1., 0.7, 0.6], [0.5, 0.1, 1.]])
    module_result, match_index, result_status = validator.verify_validity(test_numpy_array)
    assert result_status == result.ResultStatus.ALL_OKAY
    assert module_result == True


def test_validator_invalid_similarity_matrix():
    """Test verfiy validity for valid similarity matrix
    """
    test_numpy_array = numpy.array([[0.1, 0.1], [0.1, 0.1]])
    module_result, match_index, result_status = validator.verify_validity(test_numpy_array)
    assert result_status == result.ResultStatus.ALL_OKAY
    assert module_result == False



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
sys.path.insert(0,
                os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import datetime

from verifytweet import date_checker


def test_valid_date_empty_input():
    """Test for empty input in valid_date
    """
    with pytest.raises(TypeError):
        date_checker.valid_date()


def test_valid_date_invalid_type_input():
    """Test for invalid type input in valid_date
    """
    assert date_checker.valid_date(None) == False
    assert date_checker.valid_date('2018/02/23') == False
    assert date_checker.valid_date(2018) == False


def test_valid_date_invalid_input():
    """Test for invalid input in valid_date
    """
    test_date_str = 'Jan 01 1970 7:40AM'
    test_date_obj = datetime.datetime.strptime(
        test_date_str,
        '%b %d %Y %I:%M%p').replace(tzinfo=datetime.timezone.utc)
    assert date_checker.valid_date(test_date_obj) == False


def test_valid_date_valid_input():
    """Test for valid date in valid_date
    """
    test_date_obj = datetime.datetime.now(datetime.timezone.utc)
    assert date_checker.valid_date(test_date_obj) == True


def test_format_for_date_empty_input():
    """Test for empty input in format_for_date
    """
    with pytest.raises(TypeError):
        date_checker.format_for_date()


def test_format_for_date_invalid_type_input():
    """Test for invalid type input in format_for_date
    """
    with pytest.raises(TypeError):
        date_checker.format_for_date(None)
        date_checker.format_for_date('2018/02/23')
        date_checker.format_for_date(2018)


def test_format_for_date_valid_input():
    """Test for valid input in format_for_date
    """
    test_date_obj = datetime.datetime.now(datetime.timezone.utc)
    test_date_str = date_checker.format_for_date(test_date_obj)
    formatted_date_obj = datetime.datetime.strptime(
        test_date_str, '%Y-%m-%d').replace(tzinfo=datetime.timezone.utc)
    assert test_date_obj.year == formatted_date_obj.year
    assert test_date_obj.month == formatted_date_obj.month
    assert test_date_obj.day == formatted_date_obj.day



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import datetime

from verifytweet import search as search_service
from verifytweet import result

twitter_api_search = search_service.TwitterAPISearch()
twint_search = search_service.TwintSearch()


def test_search_empty_input():
    """Test search for empty input
    """
    with pytest.raises(TypeError):
        twint_search.search()


def test_search_invalid_type_input():
    """Test search for invalid type input
    """
    with pytest.raises(TypeError):
        twint_search.search(123)
        twint_search.search('123')
        twint_search.search(None, None, None)
        twint_search.search(['123'], [123], ['123'])
        twint_search.search('123', None, None)
        twint_search.search('123', '2019-07-06', 123)


def test_search_invalid_input():
    """Test search for valid type invalid input
    """
    with pytest.raises(ValueError):
        twint_search.search(
            '',
            '',
            datetime.datetime.now())


def test_search_valid_input():
    """Test search for valid input
    """
    test_user_id = 'elonmusk'
    test_tweet_snippet = 'Sharknado'
    test_tweet = 'Sharknado is real'
    module_result, module_status = twint_search.search(test_user_id,
                                                       test_tweet_snippet)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert len(module_result) > 0
    assert isinstance(module_result[0].tweet, str)
    assert test_tweet in module_result[0].tweet



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import click

from .services import controller
from .config.settings import app_config
from .util.logging import logger
from .util.result import ResultStatus


@click.command()
@click.option("--filepath",
              "-f",
              required=True,
              help="The filepath for the tweet screenshot image",
              type=click.Path(exists=True,
                              dir_okay=False,
                              resolve_path=True,
                              readable=True))
def run_as_command(filepath):
    """Verifies tweet from given image.

    Verifies tweet from image given via file path
    using a combination of image processing, text processing
    as well as a search service.

    Args:
        filepath: The filepath for the tweet screenshot image.
    
    Returns:
        Prints validity of a tweet

    """

    try:
        verify_controller = controller.NonAPIApproach()
        tweet_obj, controller_status = verify_controller.exec(filepath)
        if controller_status == ResultStatus.MODULE_FAILURE:
            print(f"Something went wrong, Please try again!")
        elif controller_status == ResultStatus.NO_RESULT:
            print(f"Fake Tweet!")
        else:
            print(f"\nVerified Tweet!")
            print(
                f"**** Username: {tweet_obj.username} ****\n**** Tweet: {tweet_obj.tweet} ****\n**** Likes: {tweet_obj.likes_count} ****\n**** Retweets: {tweet_obj.retweets_count} ****\n**** Link: {tweet_obj.link} ****"
            )
    except Exception as e:
        logger.exception(e)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import os


@pytest.fixture
def file_path():
    return os.path.abspath('./tests/static/real-tweet.png')


@pytest.fixture
def tweet_data():
    import csv

    tweet_list = list()
    with open(os.path.abspath('./tests/static/tweets.csv'),
              newline='') as csvfile:
        csvreader = csv.reader(csvfile)
        for row in csvreader:
            tweet_list.append(row[10])
    return tweet_list[1:]


@pytest.fixture
def test_data():
    import subprocess
    from verifytweet import util

    local_dir_path = os.path.abspath('./tests/static/test-dataset')
    cmd_process = subprocess.run([
        'aws', 's3', 'sync', 's3://verifytweet-dataset',
        os.path.abspath(local_dir_path)
    ])
    cmd_process.check_returncode()
    dataset = list()
    for (dirpath, dirnames, filenames) in os.walk(local_dir_path):
        if filenames:
            subset = dict()
            subset_type = dirpath.split('/')[-1]
            subset_truth = True if dirpath.split('/')[-2] == "real" else False
            file_paths = list()
            for filename in filenames:
                if util.uploader.allowed_file(filename):
                    file_paths.append(os.path.join(os.path.abspath(dirpath), filename))
            subset['expected_value'] = subset_truth
            subset['type'] = subset_type
            subset['files'] = file_paths
            dataset.append(subset)
    return dataset



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import uuid

from werkzeug.utils import secure_filename
from werkzeug.datastructures import FileStorage

from verifytweet.util.logging import logger
from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus


def save_to_disk(file_obj):
    """Saves an uploaded file via POST request to disk
    """
    if not isinstance(file_obj, FileStorage):
        raise TypeError(
            'file obj must be type werkzeug.datastructures.FileStorage')
    if not file_obj:
        raise ValueError('file obj cannot be empty')
    filename = secure_filename(file_obj.filename)
    if file_obj and allowed_file(filename):
        saved_file_name = str(uuid.uuid1()) + '.' + \
            filename.rsplit('.', 1)[1].lower()
        saved_file_path = os.path.join(app_config.FILE_DIRECTORY,
                                        saved_file_name)
        logger.debug('Saving file to path: ' + saved_file_path)
        file_obj.save(saved_file_path)
        return saved_file_path
    return None

def allowed_file(filename):
    """Checks if uploaded file has valid extenstion
    """
    return '.' in filename and \
        filename.rsplit('.', 1)[1].lower() in app_config.ALLOWED_EXTENSIONS


# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import datetime

from twint.tweet import tweet

from verifytweet import object_mapper

def test_map_keys_empty_input():
    """Test for empty input in map_keys
    """
    with pytest.raises(TypeError):
        object_mapper.map_keys()

def test_map_keys_invalid_type_input():
    """Test for invalid type input in map_keys
    """
    with pytest.raises(TypeError):
        object_mapper.map_keys(None)
        object_mapper.map_keys(dict({
            "id": "dummy",
            "tweet": "dummy"
        }))
        object_mapper.map_keys(["id", "dummy"])

def test_map_keys_invalid_input():
    """Test for valid type, invalid input in map_keys
    """
    test_tweet_obj = tweet()
    with pytest.raises(ValueError):
        object_mapper.map_keys(test_tweet_obj)

def test_map_keys_valid_input():
    """Test for valid type, valid input in map_keys
    """
    test_tweet_obj = tweet()
    test_id = "1234"
    test_username = "twitter"
    test_tweet = "Hello World!"
    test_result = dict({
        "id": test_id,
        "username": test_username,
        "tweet": test_tweet
    })
    test_tweet_obj.id = test_id
    test_tweet_obj.conversation_id = str()
    test_tweet_obj.username = test_username
    test_tweet_obj.datetime = datetime.datetime.now()
    test_tweet_obj.datestamp = str()
    test_tweet_obj.timestamp = str()
    test_tweet_obj.user_id = str()
    test_tweet_obj.name = str()
    test_tweet_obj.place = None
    test_tweet_obj.timezone = str()
    test_tweet_obj.mentions = list()
    test_tweet_obj.urls = list()
    test_tweet_obj.photos = list()
    test_tweet_obj.video = list()
    test_tweet_obj.tweet = test_tweet
    test_tweet_obj.hashtags = list()
    test_tweet_obj.replies_count = str()
    test_tweet_obj.retweets_count = str()
    test_tweet_obj.likes_count = str()
    test_tweet_obj.link = str()
    test_tweet_obj.retweet = str()
    result, module_status = object_mapper.map_keys(test_tweet_obj)
    assert result["id"] == test_result["id"]
    assert result["username"] == test_result["username"]
    assert result["tweet"] == test_result["tweet"]



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from sklearn.metrics import accuracy_score

from verifytweet import controller
from verifytweet import result

def test_overall_accuracy(test_data):
    overall_expected_output = list()
    overall_actual_output = list()
    subset_accuracy = list()
    non_api_controller = controller.NonAPIApproach()
    for subset in test_data:
        subset_expected_output = [subset['expected_value']] * len(
            subset['files'])
        overall_expected_output.extend(subset_expected_output)
        actual_output = list()
        accuracy_dict = dict()
        for file_path in subset['files']:
            module_result, module_status = non_api_controller.exec(
                file_path)
            validity = True if module_status == result.ResultStatus.ALL_OKAY else False
            actual_output.append(validity)
            overall_actual_output.append(validity)
        accuracy_dict['type'] = subset['type']
        accuracy_dict['expected'] = subset['expected_value']
        accuracy_dict['accuracy'] = accuracy_score(subset_expected_output, actual_output)
        subset_accuracy.append(accuracy_dict)
    accuracy = accuracy_score(overall_expected_output, overall_actual_output)
    print(f'Subset Accuracy: {subset_accuracy}')
    print(f'Overall Accuracy: {accuracy}')
    assert accuracy > 0.7


# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timezone

from verifytweet.config.settings import app_config
from verifytweet.util.logging import logger


def valid_date(processed_date):
    """Check if processed date of tweet is older than configured duration

    Args:
        processed_date: Represents date of processed tweet from image.

    Returns:
        A Boolean indicating if tweet can be futher processed or not. 
    """
    if not processed_date or not isinstance(processed_date, datetime):
        return False
    curr_date = datetime.now(timezone.utc)
    datetime_diff = curr_date - processed_date
    if datetime_diff.days > 7:
        return False
    return True


def format_for_date(tweet_datetime: datetime):
    if not isinstance(tweet_datetime, datetime):
        raise TypeError('Tweet date has to be type datetime')
    if not tweet_datetime:
        raise ValueError('Tweet date has to be a valid datetime object')
    return tweet_datetime.strftime('%Y-%m-%d')



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from numpy import ndarray

from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus

def verify_validity(similarity_matrix: ndarray):
    """Verifies validity of a tweet in similarity matrix.

    Verifies validity of a tweet in similarity matrix, if it crosses
    the configured threshold for similarity.

    Args:
        similarity_matrix: A list of lists containing similarity scores

    Returns:
        A Boolean representing validity of the tweet.
    """
    if not isinstance(similarity_matrix, ndarray):
        raise TypeError('Similarity matrix must type numpy.ndarray')
    row = similarity_matrix[0]
    for column_index in range(1, row.shape[0]):
        if row[column_index] > app_config.SIMILARITY_THRESHOLD:
            return (True, column_index, ResultStatus.ALL_OKAY)
    return (False, None, ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
import datetime
import twint
import urllib.parse as urlparser
import dateutil.parser as date_parser

import verifytweet.util.date_checker as date_checker

from verifytweet.util.logging import logger
from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus


class TwitterAPISearch(object):
    """Retrieve tweets of a user

    Retrieve tweets of a user tweeted on a single day
    using Twitter Search API.

    Attributes:
        user_id: A string denoting a twitter username.
        date: A datetime object representing the date in question.
    """

    def __init__(self):
        pass

    def aggregate_tweets(self, user_id: str, date: datetime.datetime):
        """Aggregates tweets from a single day.

        Retrieves tweets pertaining to the given username and date using Twitter Search API.
        Aggregates tweets to a list.

        Returns:
            A list contaning a dict representing a Tweet Object. 
            
            Ref: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/tweet-object.
            
            For example: ::

                {
                    "created_at": "Wed Oct 10 20:19:24 +0000 2018",
                    "text": "To make room for more expression, we will now count all emojis as equal—including those with gender‍‍‍ ‍‍and skin t… https://t.co/MkGjXf9aXm"
                }

        """
        if not isinstance(user_id, str) or not isinstance(
                date, datetime.datetime):
            raise TypeError(
                'User ID must be type string and date must be type datetime.datetime'
            )
        if not user_id or not date:
            raise ValueError('User ID or Date cannot be empty')
        logger.info('Searching for tweet using Twitter API...')
        querystring = dict({
            app_config.TWEET_USERNAME_KEY: user_id,
            app_config.TWEET_COUNT_KEY: app_config.TWEET_COUNT
        })
        try:
            response, response_status = self._call_twitter_api(querystring)
            if response_status != ResultStatus.ALL_OKAY:
                return (None, response_status)
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        same_day_tweets = list()
        for entry in response:
            tweet_date = date_parser.parse(entry[app_config.TWEET_DATE_KEY])
            if date_checker.format_for_date(
                    tweet_date) == date_checker.format_for_date(
                        date) and date_checker.valid_date(tweet_date):
                logger.debug('Tweet found...: ' +
                             str(entry[app_config.TWEET_TEXT_KEY]))
                same_day_tweets.append(entry)
        if not same_day_tweets:
            return (same_day_tweets, ResultStatus.NO_RESULT)
        return (same_day_tweets, ResultStatus.ALL_OKAY)

    @staticmethod
    def _call_twitter_api(querystring):
        if not isinstance(querystring, dict):
            raise TypeError('Query String must be type dict')
        if not querystring:
            raise ValueError('Query String must be a valid dictionary')
        headers = {'Authorization': 'Bearer ' + app_config.TWITTER_ACCESSTOKEN}
        search_url = urlparser.urljoin(
            app_config.TWITTER_HOSTNAME + '/' + app_config.TWITTER_APIVER +
            '/', app_config.TWITTER_CONTEXT)
        r = requests.get(search_url, headers=headers, params=querystring)
        response = r.json()
        logger.debug('Status Code for Twitter API: ' + str(r.status_code))
        if r.status_code != 200:
            raise RuntimeError('Twitter API returned status:' +
                               str(r.status_code))
        if not response:
            return (response, ResultStatus.NO_RESULT)
        return (response, ResultStatus.ALL_OKAY)


class TwintSearch(object):
    """Search for tweets using Twint

    Retrieves tweets of a user using twint basing on
    username, date as well as tweet snippet.

    Attributes:
        user_id: A string denoting a twitter username.
        date: A datetime object representing the date in question.
        tweet_snippet: A snippet of tweet extracted from image.
    """

    def __init__(self):
        pass

    def search(self, user_id: str, tweet_snippet: str,
               date: datetime.datetime = None):
        """Searches for tweets

        Retrieves tweets of given username, date as well as tweet snippet using Twint.
        Aggregates tweets to a list.

        Returns:
            A tuple contaning a list of results, each result represents a tweet object
            as well as ResultStatus.
            For example: ::

                ([<tweet_obj>], ResultStatus.ALL_OKAY)

        """
        if not isinstance(user_id, str) or not (tweet_snippet, str):
            raise TypeError(
                'User ID and tweet_snippet must be type string, date must be type datetime.datetime'
            )
        if not user_id or not tweet_snippet:
            raise ValueError('User ID, Tweet or Date cannot be empty')
        results = list()
        twint_config = twint.Config()
        twint_config.Username = user_id
        if date:
            twint_config.Since = date_checker.format_for_date(date)
            twint_config.Until = date_checker.format_for_date(date + datetime.timedelta(days=2))
        else:
            twint_config.Search = tweet_snippet
        twint_config.Limit = app_config.TWEET_MAX_STORE
        twint_config.Store_object = True
        twint_config.Store_object_tweets_list = results
        try:
            twint.run.Search(twint_config)
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        if not results:
            return (results, ResultStatus.NO_RESULT)
        logger.debug(f'Search results: {results}\n')
        return (results, ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from verifytweet import common
from verifytweet import result

def test_extract_and_parse_empty_input():
    """Test extract_and_parse for empty input
    """
    with pytest.raises(TypeError):
        common.extract_and_parse()


def test_extract_and_parse_invalid_type_input():
    """Test extract_and_parse for invalid input type
    """
    with pytest.raises(TypeError):
        common.extract_and_parse(1234)
        common.extract_and_parse(None)


def test_extract_and_parse_invalid_input():
    """Test extract_and_parse for invalid file path
    """
    with pytest.raises(ValueError):
        common.extract_and_parse('')
    module_result, result_status = common.extract_and_parse('123')
    assert result_status == result.ResultStatus.MODULE_FAILURE


def test_extract_and_parse_valid_input(file_path):
    """Test extract_and_parse for valid file path
    """
    module_result, result_status = common.extract_and_parse(file_path)
    assert result_status == result.ResultStatus.ALL_OKAY
    assert isinstance(module_result, dict)
    assert module_result['user_id'] == 'pewdiepie'
    assert module_result['tweet'] == 'ey send me stolen pdp wave designs'



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import datetime

from verifytweet import text as text_service
from verifytweet import result

data_parser = text_service.DataParser()
text_processor = text_service.TextProcessor()


def test_get_entities_empty_input():
    """Test get_entities for empty input
    """
    with pytest.raises(TypeError):
        data_parser.get_entities()


def test_get_entities_invalid_type_input():
    """Test get_entities for invalid type input
    """
    with pytest.raises(TypeError):
        data_parser.get_entities({1})
        data_parser.get_entities(None)
        data_parser.get_entities(['123'])


def test_get_entities_invalid_input():
    """Test get_entities for valid type invalid input
    """
    with pytest.raises(ValueError):
        data_parser.get_entities('')


def test_get_entities_valid_input():
    """Test get_entities for valid type valid extracted string
    """
    test_extracted_text = """ 

        Elon Musk @
        © @elonmusk CC ¥
        Ms. Tree caught the Falcon fairing!!

        1:21 AM - 25 Jun 2019

        

        2,174 Retweets 42,613 Likes oO ome C wo

        © 10K fT) 22K © 48K M4

    """
    test_result_user_id = 'elonmusk'
    test_result_tweet = 'Ms. Tree caught the Falcon fairing!!'
    test_result_datetime = datetime.datetime(2019,
                                             6,
                                             25,
                                             1,
                                             21,
                                             tzinfo=datetime.timezone.utc)
    module_result, module_status = data_parser.get_entities(
        test_extracted_text)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert test_result_user_id == module_result['user_id']
    assert test_result_tweet in module_result['tweet']
    assert test_result_datetime == module_result['date']


def test_clean_text_empty_input():
    """Test clean_text for empty input
    """
    with pytest.raises(TypeError):
        data_parser.clean_text()


def test_clean_text_invalid_type_input():
    """Test clean_text for invalid type input
    """
    with pytest.raises(TypeError):
        data_parser.clean_text(None)
        data_parser.clean_text(123)
        data_parser.clean_text([1, '2', '3'])


def test_clean_text_invalid_input():
    """Test clean_text for valid type invalid input
    """
    with pytest.raises(ValueError):
        data_parser.clean_text('')


def test_clean_text_valid_input():
    """Test clean_text for valid input
    """
    test_str = "Ms. Tree caught the Falcon fairing!!"
    module_result, module_status = data_parser.clean_text(test_str)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert module_result == "caught Falcon fairing"


def test_get_similarity_empty_input():
    """Test get_similarity for empty input
    """
    with pytest.raises(TypeError):
        text_processor.get_similarity()


def test_get_similarity_invalid_type_input():
    """Test get_similarity for invalid type input
    """
    with pytest.raises(TypeError):
        text_processor.get_similarity(123, 123)
        text_processor.get_similarity(None, None)
        text_processor.get_similarity(None, 123)
        text_processor.get_similarity([], '')


def test_get_similarity_invalid_input():
    """Test get_similarity for valid type invalid input
    """
    with pytest.raises(ValueError):
        text_processor.get_similarity('', [])


def test_get_similarity_valid_input(tweet_data):
    """Test get_similarity for valid input
    """
    from numpy import allclose, array

    test_extracted_text = "What ants must feel like pic.twitter.com/NSsBZXnEvp"
    test_result = array([[1., 0., 0., 0.09245003, 1., 0.],
                         [0., 1., 0.20412415, 0., 0., 0.],
                         [0., 0.20412415, 1., 0., 0., 0.],
                         [0.09245003, 0., 0., 1., 0.09245003, 0.],
                         [1., 0., 0., 0.09245003, 1., 0.],
                         [0., 0., 0., 0., 0., 0.]])
    module_result, module_status = text_processor.get_similarity(
        test_extracted_text, tweet_data)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert allclose(module_result, test_result)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import uuid

import PIL
import pytesseract

from verifytweet.util.logging import logger
from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus


class Extractor(object):
    """Extracts text from image

    Encapsulates image extraction process to text.

    Attributes:
        file_path: A string indicating file path where the image is stored.
    """

    def __init__(self):
        pass

    def get_text(self, file_path: str):
        """Extracts text from image
        """
        if not isinstance(file_path, str):
            raise TypeError('File path must be type string')
        if not file_path:
            raise ValueError('File path cannot be empty')
        logger.info('Processing Image...')
        try:
            new_file_path = self.rescale(file_path)
            logger.info('Extracting text from rescaled image...')
            img = PIL.Image.open(new_file_path)
            text = pytesseract.image_to_string(image=img)
            try:
                os.remove(new_file_path)
            except Exception as e:
                logger.exception(e)
            if not text:
                return (None, ResultStatus.NO_RESULT)
            return (text, ResultStatus.ALL_OKAY)
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)

    @staticmethod
    def rescale(file_path):
        if not isinstance(file_path, str):
            raise TypeError('File path must be type string')
        if not file_path:
            raise ValueError('File path cannot be empty')
        logger.info('Rescaling Image to 300 dpi...')
        new_file_path = os.path.join(app_config.FILE_DIRECTORY,
                                     str(uuid.uuid1()) + '.png')
        cmd = [
            'convert', file_path, '-resample', '300x300',
            '-alpha', 'off', '-colorspace', 'Gray', '-threshold', '75%',
            '-density', '300x300', '-units', 'PixelsPerCentimeter', '-blur',
            '1x65000', '-level', '50x100%', new_file_path
        ]
        completed_process = subprocess.run(cmd)
        completed_process.check_returncode()
        return new_file_path



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from verifytweet import image as image_service
from verifytweet import result

extractor = image_service.Extractor()

def test_get_text_empty_input():
    """Test get_text for empty input
    """
    with pytest.raises(TypeError):
        extractor.get_text()


def test_get_text_invalid_type_input():
    """Test get_text for invalid type input
    """
    with pytest.raises(TypeError):
        extractor.get_text(None)
        extractor.get_text(123)
        extractor.get_text(['123'])


def test_get_text_invalid_input():
    """Test get_text for valid type invalid input
    """
    with pytest.raises(ValueError):
        extractor.get_text('')
    assert extractor.get_text('123')[1] == result.ResultStatus.MODULE_FAILURE
    assert extractor.get_text('/home')[1] == result.ResultStatus.MODULE_FAILURE
    assert extractor.get_text('tmp.')[1] == result.ResultStatus.MODULE_FAILURE


def test_get_text_valid_input(file_path):
    """Test get_text for valid input
    """
    test_result = "ey send me stolen pdp wave designs"
    module_result, module_status = extractor.get_text(file_path)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert test_result in module_result



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twint.tweet import tweet
from verifytweet.config.settings import app_config
from verifytweet.util.logging import logger
from verifytweet.util.result import ResultStatus

def map_keys(tweet_obj):
    """Map members of tweet object to a dictionary.

    Args:
        tweet_obj: Represents a tweet object.

    Returns:
        A dictionary contaning a mapping of members of tweet object 
    """
    if not isinstance(tweet_obj, tweet):
        raise TypeError('Tweet object must be of type twint.tweet')
    try:
        id = tweet_obj.id
    except AttributeError:
        raise ValueError('Tweet object must be valid')
    return (dict({
        "id": tweet_obj.id,
        "conversation_id": tweet_obj.conversation_id,
        "datetime": tweet_obj.datetime,
        "datestamp":tweet_obj.datestamp,
        "timestamp":tweet_obj.timestamp,
        "user_id":tweet_obj.user_id,
        "username":tweet_obj.username,
        "name":tweet_obj.name,
        "place":tweet_obj.place,
        "timezone":tweet_obj.timezone,
        "mentions":tweet_obj.mentions,
        "urls":tweet_obj.urls,
        "photos":tweet_obj.photos,
        "video":tweet_obj.video,
        "tweet":tweet_obj.tweet,
        "hashtags":tweet_obj.hashtags,
        "replies_count":tweet_obj.replies_count,
        "retweets_count":tweet_obj.retweets_count,
        "likes_count":tweet_obj.likes_count,
        "link":tweet_obj.link,
        "retweet":tweet_obj.retweet
    }), ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from verifytweet.config.settings import app_config

logger = logging.getLogger()
logger.setLevel(app_config.LOG_LEVEL)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(app_config.LOG_LEVEL)

web_formatter = logging.Formatter(u'%(asctime)s -- %(levelname)s -- %(message)s')
cli_formatter = logging.Formatter(u'%(message)s')
formatter = cli_formatter if app_config.RUN_METHOD == "cli" else web_formatter

handler.setFormatter(formatter)
logger.addHandler(handler)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from werkzeug.datastructures import FileStorage

from verifytweet import uploader
from verifytweet import settings


def test_save_to_disk_empty_input():
    """Test save to disk for empty input
    """
    with pytest.raises(TypeError):
        uploader.save_to_disk()


def test_save_to_disk_invalid_type_input():
    """Test save to disk for invalid input type
    """
    with pytest.raises(TypeError):
        uploader.save_to_disk('')
        uploader.save_to_disk('123')
        uploader.save_to_disk(None)
        uploader.save_to_disk(123)


def test_save_to_disk_invalid_input():
    """Test save to disk for invalid input of valid type
    """
    test_file_obj = FileStorage('123')
    with pytest.raises(ValueError):
        uploader.save_to_disk(test_file_obj)


def test_save_to_disk_valid_input(file_path):
    """Test save to disk for valid file object
    """
    app_config = settings.app_config
    app_config.ALLOWED_EXTENSIONS = set(["png", "jpg", "jpeg"])

    with open(file_path, 'rb') as f:
        test_file_obj = FileStorage(f)
        test_file_name = uploader.save_to_disk(test_file_obj)
    assert isinstance(test_file_name, str)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import re

from setuptools import find_packages
from setuptools import setup

with io.open("README.md", "rt", encoding="utf8") as f:
    readme = f.read()

with io.open("verifytweet/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setup(
    name="verifytweet",
    version=version,
    url="https://preethamkamidi.com/projects/verify",
    project_urls={
        "Documentation": "https://github.com/kamidipreetham/verifytweet",
        "Code": "https://github.com/kamidipreetham/verifytweet",
        "Issue tracker":
        "https://github.com/kamidipreetham/verifytweet/issues",
    },
    license="AGPLv3",
    author="Preetham Kamidi",
    author_email="contact@preethamkamidi.com",
    description="A tool to verify Tweet screenshots",
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6.*",
    install_requires=[
        "click>=5.1", "Pillow>=6.2.0", "pytesseract==0.2.6",
        "requests==2.22.0", "scikit-learn==0.21.2", "nltk>=3.4.5",
        "python-dateutil==2.8.0", "werkzeug==0.15.4",
        "twint==2.1.13"
    ],
    entry_points={
        "console_scripts": ["verifytweet = verifytweet.cli:run_as_command"]
    },
)


# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from flask import Flask, jsonify, request
from flask_cors import CORS

from .services import controller
from .services import image as image_service
from .util import uploader as image_uploader
from .util import object_mapper
from .util.logging import logger
from .config.settings import app_config
from .util.result import ResultStatus

router = Flask(__name__, static_folder=app_config.FILE_DIRECTORY)
router.config['MAX_CONTENT_LENGTH'] = app_config.MAX_CONTENT_LENGTH
cors = CORS(router, resources={r'/api/*': {'origins': '*'}})


@router.route('/', methods=['GET'])
def root():
    logger.info('Health Check log')
    return jsonify({'status': 0})


@router.route('/health', methods=['GET'])
def health():
    logger.info('Health Check log')
    return jsonify({'status': 0})


@router.route('/api/v1/verify', methods=['POST'])
def verify_tweet():
    """Verifies tweet from given image.

    Verifies tweet from image received via POST request
    using a combination of image processing, text processing
    as well as a search service.

    Args:
        data: File object containing Image data.
        type: A string indicating the type of data: "image" or "link".

    Returns:
        A dict mapping status and validity of the tweet. For
        example: ::

            {
                "status": 0,
                "result": True
            }
    
        If the form fields are missing or the tweet is too old,
        a 400 (Bad Request) is returned.

    """
    logger.info('Received data for processing...')

    data_type = request.form['type']
    request_image = request.files['data']
    if not data_type or not request_image:
        return "Missing form fields", 400
    try:
        file_path = image_uploader.save_to_disk(request_image)
        rest_controller = controller.NonAPIApproach()
        result, controller_status = rest_controller.exec(file_path)
    except Exception as e:
        logger.exception(e)
        return jsonify({
            'status': ResultStatus.MODULE_FAILURE.value,
            'result': None
        })
    if controller_status != ResultStatus.ALL_OKAY:
        return jsonify({'status': controller_status.value, 'result': result})
    tweet_dict, mapper_status = object_mapper.map_keys(result)
    if mapper_status != ResultStatus.ALL_OKAY:
        return jsonify({'status': mapper_status.value, 'result': tweet_dict})
    return jsonify({'status': mapper_status.value, 'result': tweet_dict})



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import datetime
import string
import random
import nltk
import dateutil.parser as date_parser

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity

from verifytweet.util.logging import logger
from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus

nltk.download('stopwords', quiet=True)
nltk.download('punkt', quiet=True)
count_vectorizer = CountVectorizer()
stopwords = set(nltk.corpus.stopwords.words('english'))

USERNAME_REGEX = r'@(\w{1,15})\b'
DATETIME_REGEX = r'((1[0-2]|0?[1-9]):([0-5][0-9]) ?([AaPp][Mm]))\s-\s\d{1,2}\s\w+\s\d{4}'
ALPHANUM_REGEX = r'[^A-Za-z0-9]+'


class DataParser(object):
    """Parses data from extracted text
    """

    def __init__(self):
        pass

    def get_entities(self, extracted_text: str):
        """Parses entities from extracted text.

        Parses username (denoted by user_id), tweet as well as date from extracted text.

        Attributes:
            extracted_text: A string denoting extracted text from image.

        Returns:
            A tuple contaning a dictionary: a mapping of user_id, tweet and date
            as well as Enum ResultStatus which gives out result status.
            For example: ::

                {
                    "user_id": "elonmusk",
                    "tweet": "Ms. Tree caught the Falcon fairing!!",
                    "date": datetime.datetime(2019, 6, 8, 7, 29, tzinfo=datetime.timezone.utc)
                }

        """
        if not isinstance(extracted_text, str):
            raise TypeError('Extracted text must be type string')
        if not extracted_text:
            raise ValueError('Extracted text cannot be empty')
        logger.info('Parsing data out of extracted text...')
        username_match = re.search(USERNAME_REGEX, extracted_text)
        datetime_match = re.search(DATETIME_REGEX, extracted_text)
        if not username_match:
            return (dict({
                'user_id': None,
                'tweet': None,
                'datetime': None
            }), ResultStatus.NO_RESULT)
        user_id = username_match.group()[1:]
        tweet_start_index = username_match.end()
        tweet_end_index = len(
            extracted_text
        ) - 1 if not datetime_match else datetime_match.start()
        tweet = extracted_text[tweet_start_index:tweet_end_index].strip()
        if not datetime_match:
            return (dict({
                'user_id': user_id,
                'tweet': tweet,
                'date': None
            }), ResultStatus.ALL_OKAY)
        date_str = datetime_match.group().replace('-', '')
        processed_datetime = date_parser.parse(date_str).replace(
            tzinfo=datetime.timezone.utc)
        return (dict({
            'user_id': user_id,
            'tweet': tweet,
            'date': processed_datetime
        }), ResultStatus.ALL_OKAY)

    def clean_text(self, extracted_text: str):
        """Removes stop words and samples words out of tweet
        to create a snippet.

        Attributes:
            extracted_text: A string denoting extracted text from image.

        Returns:
            A tuple contaning a tweet snippet
            as well as Enum ResultStatus which gives out result status.
        """
        if not isinstance(extracted_text, str):
            raise TypeError('Extracted text must be type string')
        if not extracted_text:
            raise ValueError('Extracted text cannot be empty')
        try:
            non_punc_tweet = extracted_text.translate(
                str.maketrans('', '', string.punctuation))
            word_tokens = nltk.tokenize.word_tokenize(non_punc_tweet)
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        filtered_sentence = [w for w in word_tokens if not w in stopwords]
        picked_words = filtered_sentence[2:min([len(filtered_sentence), 6])]
        tweet_snippet = " ".join(picked_words)
        if not tweet_snippet:
            return (tweet_snippet, ResultStatus.NO_RESULT)
        logger.debug(f'Tweet Snippet: {tweet_snippet}')
        return (tweet_snippet, ResultStatus.ALL_OKAY)


class TextProcessor(object):
    """Processes extracted tweet and aggregated tweets
    """

    def __init__(self):
        pass

    def get_similarity(self, extracted_tweet: str, same_day_tweets: list):
        """Calculates a similarity matrix.

        Calculates a similarity matrix of the corpus containing
        extracted tweet and tweets aggregated from Twitter Search API
        using consine similarity approach.

        Attributes:
            extracted_tweet: A string denoting extracted tweet from image.
            same_day_tweets: A list contaning tweets of target date

        Returns:
            A tuple contaning a similarity matrix, which is a numpy array
            as well as Enum ResultStatus which gives out result status.
            For example: ::

                ([[1.        0.9258201]
                 [0.9258201 1.       ]], ResultStatus.ALL_OKAY)


        """
        if not isinstance(extracted_tweet, str) or not isinstance(
                same_day_tweets, list):
            raise TypeError(
                'Extracted tweet must be type str and Same day tweets must be type list'
            )
        if not extracted_tweet or not same_day_tweets:
            raise ValueError(
                'Extracted tweet must be a valid string and same day tweets must be a valid list'
            )
        logger.info('Processing similarity of two tweets...')
        corpus = list()
        corpus.append(extracted_tweet)
        corpus.extend(same_day_tweets)
        logger.debug('Corpus: ' + str(corpus))
        try:
            sparse_matrix = count_vectorizer.fit_transform(corpus)
            similarity_matrix = cosine_similarity(sparse_matrix, sparse_matrix)
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        logger.debug('Similartiy Matrix: ' + str(similarity_matrix))
        return (similarity_matrix, ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import verifytweet.services.image as image_service
import verifytweet.services.text as text_service
import verifytweet.services.search as search_service
import verifytweet.util.date_checker as date_checker
import verifytweet.util.common as common

from verifytweet.util.logging import logger
from verifytweet.config.settings import app_config
from verifytweet.util.result import ResultStatus


class APIApproach(object):
    """Use Twitter API to verify tweet
    """

    def __init__(self):
        pass

    def exec(self, file_path: str):
        """Executes controller flow

        Controller uses image service to extract text from
        image, passes text to text service to parse entities such
        as username, tweet as well as date, uses search service 
        to retrieve same day tweets, text service to find similar tweet
        and finally verifying the tweet.

        Attributes:
            file_path: A string denoting a twitter username.

        Returns:
            valid_tweet: A tweet object
            status: Enum ResultStatus representing result status

        """
        if not isinstance(file_path, str):
            raise TypeError('File path must be type str')
        if not file_path:
            raise ValueError('File path must be a valid string')
        entities, preprocess_status = common.extract_and_parse(file_path)
        if preprocess_status != ResultStatus.ALL_OKAY:
            return (None, ResultStatus.MODULE_FAILURE)

        try:
            search_controller = search_service.TwitterAPISearch()
            same_day_tweets, search_status = search_controller.aggregate_tweets(
                entities['user_id'], entities['date'])
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        if search_status != ResultStatus.ALL_OKAY:
            return (None, search_status)
        validity, match_index, validator_status = common.calculate_and_validate(
            entities=entities, same_day_tweets=same_day_tweets)
        if validator_status != ResultStatus.ALL_OKAY:
            return (None, ResultStatus.MODULE_FAILURE)
        return (same_day_tweets[match_index], ResultStatus.ALL_OKAY)


class NonAPIApproach(object):
    """Use a non-api approach to verify tweet
    """

    def __init__(self):
        pass

    def exec(self, file_path):
        """Executes controller flow

        Attributes:
            file_path: A string denoting a twitter username.

        Controller uses image service to extract text from
        image, passes text to text service to parse entities such
        as username, tweet as well as date, uses search service 
        to retrieve a tweet if available.

        Returns:
            search_results: list of tweet objects
            status: Enum ResultStatus representing result status

        """
        if not isinstance(file_path, str):
            raise TypeError('File path must be type str')
        if not file_path:
            raise ValueError('File path must be a valid string')
        entities, preprocess_status = common.extract_and_parse(file_path)
        if preprocess_status != ResultStatus.ALL_OKAY:
            return (None, ResultStatus.MODULE_FAILURE)

        try:
            text_processor = text_service.DataParser()
            tweet_snippet, text_processor_status = text_processor.clean_text(
                entities['tweet'])
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        if text_processor_status != ResultStatus.ALL_OKAY:
            return (None, text_processor_status)

        try:
            search_controller = search_service.TwintSearch()
            search_results, search_status = search_controller.search(
                entities['user_id'], tweet_snippet, entities['date'])
        except Exception as e:
            logger.exception(e)
            return (None, ResultStatus.MODULE_FAILURE)
        if search_status != ResultStatus.ALL_OKAY:
            return (None, search_status)
        tweet_text_list = list()
        for tweet_obj in search_results:
            tweet_text_list.append(tweet_obj.tweet)
        validity, match_index, validator_status = common.calculate_and_validate(
            entities=entities, tweet_text_list=tweet_text_list)
        if validator_status == ResultStatus.MODULE_FAILURE:
            return (None, ResultStatus.MODULE_FAILURE)
        if not validity:
            return (None, ResultStatus.NO_RESULT)
        return (search_results[match_index], ResultStatus.ALL_OKAY)



# Verify Tweet verifies tweets of a public user
# from tweet screenshots: real or generated from
# tweet generators.
# Copyright (C) 2019 Preetham Kamidi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from verifytweet import controller
from verifytweet import result

non_api_approach = controller.NonAPIApproach()


def test_exec_empty_input():
    """Test exec for empty input
    """
    with pytest.raises(TypeError):
        non_api_approach.exec()


def test_exec_invalid_type_input():
    """Test exec for invalid type input
    """
    with pytest.raises(TypeError):
        non_api_approach.exec(None)
        non_api_approach.exec(123)
        non_api_approach.exec(['/home/'])


def test_exec_invalid_input():
    """Test exec for valid type invalid input
    """
    with pytest.raises(ValueError):
        non_api_approach.exec('')
    assert non_api_approach.exec(
        '123')[1] == result.ResultStatus.MODULE_FAILURE
    assert non_api_approach.exec(
        '/home')[1] == result.ResultStatus.MODULE_FAILURE
    assert non_api_approach.exec(
        'tmp.png')[1] == result.ResultStatus.MODULE_FAILURE


def test_exec_valid_input(file_path):
    """Test exec for valid input
    """
    from twint.tweet import tweet

    test_result_tweet = 'ey send me stolen pdp wave designs'
    test_result_username = 'pewdiepie'
    module_result, module_status = non_api_approach.exec(file_path)
    assert module_status == result.ResultStatus.ALL_OKAY
    assert isinstance(module_result, tweet)
    assert test_result_tweet in module_result.tweet
    assert test_result_username == module_result.username
